from __future__ import annotations

from typing import Optional

import numpy as np
from scipy.optimize import fsolve

try:
    from scipy.integrate import cumulative_trapezoid
except ImportError:
    from scipy.integrate import cumtrapz as cumulative_trapezoid
from scipy.stats import linregress

from .points import Points

from ..dsc.tnmh_model import TNMH


class DSC(Points):
    def __init__(self, x, y, **kwargs):

        y = np.asarray(y).reshape(np.asarray(x).shape)

        x, unique = np.unique(x, return_index=True)

        super().__init__(x, y[unique], **kwargs)

    def get_fictive_cp(self, glass: tuple[float, float], liquid: tuple[float, float]) -> ('DSC', float):
        min_glass, max_glass = min(glass), max(glass)
        min_liquid, max_liquid = min(liquid), max(liquid)

        region = self.copy()
        region.cut(min_glass, max_liquid)

        glass_regime = (min_glass < region.x) & (region.x < max_glass)
        regress = linregress(region.x[glass_regime], region.y[glass_regime])
        glass_extrapolation = regress.slope * region.x + regress.intercept

        region.y -= glass_extrapolation

        liquid_regime = (min_liquid < region.x) & (region.x < max_liquid)
        regress2 = linregress(region.x[liquid_regime], region.y[liquid_regime])

        real_area = cumulative_trapezoid(region.y, region.x, initial=0)
        real_area -= real_area[-1]

        t = regress2.intercept-regress.intercept
        m = regress2.slope-regress.slope
        c0 = 0.5*m*region.x.max()**2 + t*region.x.max()

        def equiv(_x, _i):
            return (0.5*m * _x**2 + t*_x - c0) - real_area[_i]

        def equiv_prime(_x, _i):
            return m * _x + t

        fictive_temperature = np.array([fsolve(equiv, region.x[i], fprime=equiv_prime, args=(i,))[0] for i in range(len(region.x))])

        t_g_fictive = fictive_temperature[:20].mean()
        region.y = np.gradient(fictive_temperature, region.x)

        return region, t_g_fictive

    def calculate_tnmh(self, p0: list, glass: tuple[float, float], liquid: tuple[float, float],
                       tg: float = None, num_points: int = 200, return_fictive: bool = True) -> ('FitResult', Optional[float], Optional[DSC]):

        dtf_dt, fictive_tg = self.get_fictive_cp(glass, liquid)
        if tg is None:
            tg = fictive_tg

        temp_equidist = np.linspace(dtf_dt.x[0], dtf_dt.x[-1], num_points)
        dtf_dt_equidist = np.interp(temp_equidist, dtf_dt.x, dtf_dt.y)

        from ..fit.minimizer import FitRoutine
        fitter = FitRoutine()
        fitter.set_model(TNMH)
        data = fitter.add_data(temp_equidist, dtf_dt_equidist)
        data.set_parameter(p0 + [tg, self.value], var=[True]*4 + [False]*2, default_bounds=True)

        res = fitter.run()[0]

        if return_fictive:
            return res, tg, dtf_dt
        else:
            return res
